`parse_msg_file` <- function(subjectcode, msgdir = './msg/')
{
#-------------------------------------------------------------------
#  FUNCTION parse_msg_file_l3.m
#  by Jochen Laubrock
#  example parser for SR resarch EyeLink message file
#  (this example from "lesen III" with PSC2)
#  - parse message file, return relevant info in tab
#		note: because sentence info (start on screen etc.) was not written to the edf file, 
#	 	we need to load it from an external file and match it to tab by sentence-ID. 
#-------------------------------------------------------------------
#  INPUT:
#
#  data from lesen III (psc2) experiment, 
#  Eyelink 2 system, Gutenbergstr, 2006
#
#  subjectcode          subject code
#  *.msg file   message file
#  msgdir            path to file 
#  
#  OUTPUT:
#  matrix 'tab' where each line is a trial = sentence
#  tab(1:ntrials,1)   sentence identification number
#  tab(1:ntrials,2)   x-coordinate of sentence beginning
#  tab(1:ntrials,3)   y-coordinate of sentence beginning (should always be ca. 316 Pixel)
#  tab(1:ntrials,4)   start time of sentence presentation
#  tab(1:ntrials,5)   end time of sentence presentation
#  tab(1:ntrials,6)   x-coordinate of sentence end
#---------------------------------------------------------------------

# # lesen III 
# # file header
#
# MSG	32944 BEGIN OF DESCRIPTIONS
# MSG	32944 VP-Nr. PSCII001						# SUBJECT-ID PSCII001
# MSG	32944 SCREEN_COORDS 0 0 1023 767
# MSG	32945 SCREEN_INFO MO_FREQ BGCOLOR 120 255
# MSG	32945 TEXTLINE 384  PIXPERLETTER
# MSG	32945 SUBJECTDISTANCE 600 in mm
# MSG	32945 MONITORWIDTH 400 in mm
# MSG	32945 END OF DESCRIPTIONS
#
# # per sentence
#
# MSG	791160 SYNCTIME
# MSG	791035 TRIALID 26 SATZ 73		# TRIALID 26 SENTENCE 73
# MSG	791160 TRIAL_START 26			# TRIAL_START
# MSG	791704 SATZ_EINGEBLENDET 		# SENTENCE_PRESENTED
# MSG	793093 TRIAL_ENDE 26			# TRIAL_END


# for testing:
# sub{1}='p2or1';
# msgdir <- "/Users/jochen/Projects/work/eyelinkTest/lesen\ III/data/asc/psc2/"
# filename <- "PSCII001.msg2"

# # parms.R is a text file containing info on PIX_PER_CHAR etc.
# libpath <- '/Users/jochen/Projects/R\ library/eyetrackR/msgparser'
# source(paste(libpath, 'parms.R', sep='/'))
# source('parms.R')

# source('readParms.R')
# parmsfile <- 'parms.dat'
# readParms(parmsfile)

data(parms1)

# load(corpusinfofile) # contains variable PSC2.DWDS with corpus info (used to be called PSC2.DWDSnew.rda)
# corpusinfofile <- 'PSC2.DWDS'
# data(corpusinfofile) # contains variable PSC2.DWDS with corpus info
data('PSC2.DWDS')

# filename = sprintf('%s.msg2', subjectcode);    % file name  
# fprintf(1,'\nsubject %s\n', subjectcode);
filename = sprintf('%s.msg2',subjectcode);    # file name  
cat(paste('\nsubject', subjectcode, '\n'))
cat(paste('\t... pre-processing message file', filename, '\n'));   # write formatted data to file

myfile <- readLines(con=paste(msgdir,filename,sep=''))

monitor <- list(coords = rep(NA,4), hz=NA)

tmp <- myfile[3] # MSG	32944 SCREEN_COORDS 0 0 1023 767
tmp <- strsplit(tmp, '[[:blank:]]')[[1]]
if (tmp[3] == "SCREEN_COORDS") {     monitor$coords <- as.double(tmp[4:7]) } 
else { stop('third line has wrong format') }

tmp <- myfile[4] # MSG	32945 SCREEN_INFO MO_FREQ BGCOLOR 120 255
tmp <- strsplit(tmp, '[[:blank:]]')[[1]]
if (tmp[3] == "SCREEN_INFO") { monitor$hz <- as.double(tmp[6]) } 
else { stop('fourth line has wrong format') }

# indices of all TRIALID messages
alltrialidIx <- grep('TRIALID', myfile)

# # indices of all TRIAL_START or TRIAL_STARTWITHIN messages
alltrialstartIx <- grep('(TRIAL_START)|(TRIAL_STARTWITHIN)', myfile)

# indices of all SATZ_EINGEBLENDET messages
alltextonIx <- grep('SATZ_EINGEBLENDET', myfile) 

# use only the last TRIAL_START(_WITHIN) message preceding TEXT_ON 
tmp <- rep(NA, length(alltextonIx))
for (i in seq(along=alltextonIx)) {
	tmp[i] <- alltrialstartIx[ max(which(alltrialstartIx < alltextonIx[i])) ]
}
alltrialstartIx <- tmp

# indices of all TRIAL_ENDE messages
alltrialendIx <- grep('TRIAL_ENDE', myfile)

# delete indices of training trials (trialnumber < 1)
tmp <- grep('TRIALID ((-)|(0))', myfile[alltrialidIx])
exptrialIx <- setdiff(1:length(alltrialidIx), tmp)

alltrialidIx <- alltrialidIx[exptrialIx]
alltrialstartIx <- alltrialstartIx[exptrialIx]
alltextonIx <- alltextonIx[exptrialIx]
alltrialendIx <- alltrialendIx[exptrialIx]

n1 <- length(alltrialidIx)
n2 <- length(alltrialendIx) 
if (n1!=144 || n2!=144){
    if (n1!=n2){
        stop('strange, different number of trial start and end messages!!!')
	} else { 
		cat(paste(n1, 'out of 144 sentences valid\n'))
	}
}

# indices of responses ( i think we didn't have responses here)
# respIx <- grep('RESPONSE', myfile)

S <- list(trialNo=NULL, sentenceNo=NULL, sentence=NULL, wordBoundaries=NULL, textbox=NULL, correctResponse=NULL, t0=NULL, t1=NULL)
x0 <- rep(NA, n1) 
xend <- rep(NA, n1) 
y0 <- rep(NA, n1) 

# take 4th and 6th field of messages
# MSG	483613 TRIALID 1 SATZ 102
# MSG	487731 TRIALID 2 SATZ 131
# ...
trialIdRows <- strsplit(myfile[alltrialidIx], '[[:blank:]]')
trialNo <- as.double(unlist(trialIdRows)[seq(4, 6*n1, by=6)]) # see below (next line) for a possibly nicer alternative
sID <- as.double(sapply(trialIdRows, '[', 6)) # sweet, isn't it 
# (in long): apply function [ with argument 6 to the list to extraxt trialIdRows[[i]][6] from each item of list trialIdRows
if (length(trialNo) != length(sID)) {stop('length(trialNo) != length(sID)')}

# MSG	483748 TRIAL_START 1
trialStartRows <- strsplit(myfile[alltrialstartIx], '[[:blank:]]')
tstart <- as.double(sapply(trialStartRows, '[', 2)) # sweet, isn't it

# MSG	484073 SATZ_EINGEBLENDET 
sentenceStartRows <- strsplit(myfile[alltextonIx], '[[:blank:]]')
sentenceTimestamp <- as.double(sapply(sentenceStartRows, '[', 2)) 
sentenceTime <- sentenceTimestamp - tstart

# MSG	1920078 TRIAL_ENDE 142
trialEndRows <- strsplit(myfile[alltrialendIx], '[[:blank:]]')
trialendTimestamp <- as.double(sapply(trialEndRows, '[', 2)) 
trialendTrial <- as.double(sapply(trialEndRows, '[', 4))
if (length(unique(trialendTrial == trialNo))==2 ||  !unique(trialendTrial == trialNo)) {
	errlist <- which(trialendTrial != trialNo)
	cat('problems in trial(s):', errlist, '\n')
	stop('parse_msg_file: no matching trial_ende for trials')
}

x0 <- rep(40, n1) # this is a constant in eike-style experiments
y0 <- rep((monitor$coords[4]+1)/2, n1); # assumption, better to measure!!
x1 <- rep(NA, n1) # this is a constant in eike-style experiments
for (i in 1:n1) {
	ix <- PSC2.DWDS$sn == sID[i]
	x1[i] <- x0[i] + (sum(PSC2.DWDS[ix,"nlett1"]) + sum(ix)-1) * PIX_PER_CHAR # offset + (sum(letters) + N spaces) * ppc
}

tab <- as.data.frame(cbind(sID, x0, y0, x1, sentenceTimestamp, trialendTimestamp))
names(tab) <- c("sID", "x0", "y0", "x1", "sentenceTimestamp", "trialendTimestamp")

tab
}