# =======================================================
#  PROGRAM xallsac_l3.R                               
#  by jochen laubrock                                 
#  cleaned-up and simplified version of antjes stuff  
#  for lesen III (based on lautesLesen analysis)      
#  ported to R
# =======================================================
#
# a real-life example of how to get from raw data to a fixation sequence
#
# The file is NOT meant to be executed with demo(xallsac), but rather 
# to be studied and used as a starting point for own analyses programs
#
#
# still needs further cleanup
# TODO 
# write all tables only at the end
# think about how to make this more generic (provide an interface)
#-------------------------------------------------------------------
#
#  INPUT:           data from reading experiment:
#                   *.msg2          for a given subject, text file with messages regarding the experiment 
#									(generated by piping msg file through a few grep expressions--we will provide an R script to do this, too)
#                   *.dat           for a given subject, file with raw data (here: 7 columns, but could also be 4--depends on tracker)
#                   PSC2_DWDS.mat   corpus-level information on the Potsdam Sentence Corpus (PSC2)
#
#  OUTPUT:          .seq file containing sequence of fixations
#
#  FORMAT:          every row is a fixation
#                   16 columns
#                   1) subject id,
#                   2) sentence id,
#                   3) number of words
#                   4) word number (fixated word)
#                   5) letter number (fixated letter),
#                   6) fixation duration [ms]
#                   7) duration of the following saccade [ms]
#                   8) peak velocity [deg/ sec]
#                   9) dX x amplitude of the following saccade [deg]
#                   10) dY y amplitude of the following saccade [deg]
#                   11) dx x horizontal component
#                   12) dy y vertical component
#                   13) x position of gaze, in screen coordinates (pixels) at beginning of fixation, fpos_a(k)
#                   14) x position of gaze, in screen coordinates (pixels) at end of fixation, fpos_b(k)
#                   15) x position of gaze, in screen coordinates (pixels), averaged across fixation, fpos(k)
#                   16) first (1) vs. last (2) fixation within a sentence (0 otherwise)
#
#  dependencies:  readSubjects.R, parse_msg_file.R, getMissingData.R, vecvel.R, microsacc.R, binsacc.R, wordLetterPos.R, treatBadData.R
#---------------------------------------------------------------------
# screen resolution during (this) reading experiment: 1024 x 768 Pixel
# distance between screen monitor and subject: 60 cm
#
# NB: my recommendation: in reading, always use binsacc 

rm(list=ls())
library(eyetrackR)
# provides 'readSubjects.R', 'parms1.R', 'parse_msg_file.R', 'treatBadData.R', 'wordLetterPos.R'

# =============================================================================
# =                                                                           =
# =                              PARAMETER BLOCK                              =
# =                (adapt the following lines to your needs)                  =
# =                                                                           =
# =============================================================================

exp_str <- 'PSCII'   		# string characterizing the experiment

# ===========================
# load corpus
# ===========================
data(parms1)
data(PSC2.DWDS) # contains variable PSC2.DWDS with corpus info
# alternatively:
# source('Full_Path_To_My_Parms_File.R')
# load('Full_Path_To_My_Corpus_File.rda')

# ===========================
# = define some directories =
# ===========================
basepath <- '~/Projects/work/eyelinkTest/lesen III'
projectpath <- file.path(basepath, 'neueAuswertung/SacDetec2008/PSC2')
# paths for raw input data
basedatapath <- file.path(basepath, 'data/psc2/asc')
msgfilepath <- file.path(basedatapath, 'msg/')
datfilepath <- file.path(basedatapath, 'dat/')
# where the sequence files will be stored
outputpath <- file.path(projectpath, 'fix_seqs_with_R') 
# where info on missing and/or 'bad' data will be saved
missingpath <- file.path(outputpath, 'missings') 

# =============================================================================
# =                                                                           =
# =                            END PARAMETER BLOCK                            =
# =                                                                           =
# =============================================================================

if (!file_test("-d", basepath)) {stop('please correct basepath')}
if (!file_test("-d", msgfilepath)) {stop('please correct msgfilepath')}
if (!file_test("-d", datfilepath)) {stop('please correct datfilepath')}
if (!file_test("-d", outputpath)){dir.create(outputpath)}
if (!file_test("-d", missingpath)){dir.create(missingpath)}


print('=====================================================================')
print('This file is NOT meant to be executed with demo(xallsac), but rather')
print('to be studied and used as a starting point for own analyses programs')
print('=====================================================================')

print('\n\n+++ conversion of raw data into fixation sequences +++\n');


# Read subject list (and check for availability of associated raw data sets)
subjects <- readSubjects(exp_str, dir=basedatapath, msgdir=msgfilepath, datdir=datfilepath)
mis_collect <- NULL

# =======================
# = prepare output files 
# =======================

sacdefLbls <- c('LEFT', 'RIGHT', 'BOTH')
whichsacs <- c('sacl', 'sacr', 'sacb')
outfilename <- rep(NA, length(sacdefLbls))
outfiles <- list()

nos <- length(subjects) # number of subjects
# construct file names and open output file for fixation sequence data

basename <- file.path(outputpath, paste(exp_str, '_allsac_all_', sep=''))
for (lbl in sacdefLbls) {
		outfilename[lbl] <- paste(basename, lbl,'.seq', sep='')
		outfiles[[lbl]] <-  file(outfilename[lbl], open = 'w');
}

# collection matrices for 'missing data' (same for both eyes)
mis_col_s_trial_id <- NULL
mis_col_NrPerSubject <- NULL

# collection matrices for 'bad data' (same for both eyes)
mis2_col_s_trial_id <- NULL
mis2_col_NrPerSubject <- NULL

cat('\nPARAMETERS FOR (MICRO)SACCADE DETECTION:\n');
cat(sprintf('velocity type %i, velocity threshold i, minimum duration of saccade %i samples\n', VTYPE, VTHR, MINDUR))
cat(sprintf('sampling rate %i, PPC %i', SAMPLING, PIX_PER_CHAR))

# ======================
# = loop over subjects =
# ======================

for (s in 1:nos) {  # loop over subjects =
	sub <- subjects[s] # Create character array (string)
	
	subjID <- as.numeric(strsplit(sub, exp_str)[[1]][2]) # needs to be adapted according to how output file names are constructed
	
	# ====================================================
	# 1)    get relevant information from message file
	# 		(relies on function parse_msg_file, which
	# 		needs to be adapted to each new experiment)
	# ====================================================
	tab <- parse_msg_file(sub,msgfilepath)   # call function parse_msg_file to parse message file 
	
	cat(sprintf('\n\t... loading raw data for %s (No. %i, subject %i) \n', sub, s, subjID))  
	datfile <- sprintf('%s.dat',sub)
	data <- read.table(paste(datfilepath, datfile, sep='/'), nrows=800000, col.names=c('t', 'xl', 'yl', 'pl', 'xr', 'yr', 'pr'))   # load file with raw data
	# !!!! note: the nrows default in the above is dangerous: if file is longer,  it is truncated,  which leads to "no samples found " messages for trials late in the experiment
	# however,  without providing nrows,  read.table is very slow
	# make sure to use a generous estimate
	# TODO: maybe it's possible to guesstimate from file size?
	
	cat(sprintf('\t... raw data for %s (subject %i) are converted into fixation sequences\n', sub, s))
	ttime <- data$t # time stamp from 1. column of raw data matrix
	for (trial in 1:nrow(tab))  {# loop over trials
		id <- tab[trial, 'sID']      	# get sentence ID 
		sentenceInCorpusIx <- PSC2.DWDS$sn==id # find sentence in corpus
		nwords <- PSC2.DWDS[sentenceInCorpusIx, 'nw'][1]   # number of words with sentence 'id'
		sentenceWL <- PSC2.DWDS[sentenceInCorpusIx, "nlett1"]
		
		# find rows representing the current trial and/ or sentence (logical index)
		indx <- ttime >= tab[trial, "sentenceTimestamp"] & ttime <= tab[trial, "trialendTimestamp"]
		# treat missing-data trials
		if (sum(indx)==0) {
			cat('no samples found for subject', sub, 'trial',trial,'\n'); # print error message on screen if no data points found
			mis2_col_s_trial_id <- rbind(mis2_col_s_trial_id, c(s, trial, id))    # collect them
			next # skip over remainder of trial loop
		}
		time <- ttime[indx]  # consider only relevant time stamps
		
		# ====================================================
		# 2)    Detection of missing values
		# ====================================================
		# WE NEED TO REPLACE THIS HACK (ASSUMPTION OF FIXED SENTENCE START AT 40 PIX) WITH EMPIRICAL MEASUREMENT
		if (is.nan(tab[trial, "x1"])) {
			# from offset, sum up all letters, add one space per word, subtract one space overall, multiply by PPC
			offset <- LEFT_SENTENCE_OFFSET         # x-coordinate for sentence onset
			rightSentenceBorder <- LEFT_SENTENCE_OFFSET + (sum(PSC2.DWDS[sentenceInCorpusIx, "nlett1"] + 1) - 1) * PIX_PER_CHAR
		} else {
			offset <- tab[trial,"x0"]                  # x-coordinate for sentence onset
			rightSentenceBorder <- tab[trial, "x1"]
			rs1 <- LEFT_SENTENCE_OFFSET + (sum(PSC2.DWDS[sentenceInCorpusIx, "nlett1"] + 1)  - 1)* PIX_PER_CHAR  # add one space per word, subtract one space overall
			#        rightSentenceBorder = sentenceBorder(id,satz,PIX_PER_CHAR);    # relative to sentence beginning!
			if (rs1 != rightSentenceBorder) {
				cat(sprintf('different corpus? check vp %s trial %d sentence %d\n', sub, trial, id))
			}
		}
		data_trial <- data[indx,]
		# define missing data, ignore missings after right sentence border
		res <- getMissingData(data_trial, rightSentenceBorder, PIX_PER_CHAR)
		missing <- res$missing
		data_trial <- res$data_trial
		
		if (length(missing)>0)  {  # if there are missings, then collect them
			mis_col_s_trial_id = rbind(mis_col_s_trial_id, c(s, trial, id))  # subject id, trial no., sentence no.
		} else { # if there are no missings with this trial, then continue ...
			# ====================================================
			# 3)    Detection of all saccades
			# ====================================================
			xyl <- DEG_PER_PIX * data_trial[, c("xl", "yl")] # x- and y-position for LEFT eye
			vl <- vecvel(xyl, SAMPLING, VTYPE);       # computation of velocity components (x & y) for LEFT eye data
			sacl <- microsacc(xyl, vl, VTHR, MINDUR);  # LEFT eye,output: '7 x number of saccades'-matrix
			
			xyr <- DEG_PER_PIX * data_trial[, c("xr", "yr")] # x- and y-position for RIGHT eye
			vr <- vecvel(xyr, SAMPLING, VTYPE);        # computation of velocity components (x & y) for RIGHT eye data
			sacr <- microsacc(xyr, vr, VTHR, MINDUR);  # RIGHT eye,output: '7 x number of saccades'-matrix
			
			# binocularity criterion
			# 	note: blame me, I reintroduced this after inspecting the raw data:
			# 	a pure velocity-based criterion simply produced way too many
			# 	ultra-short fixations--there is a lot of under-/overshoot (jl) 
			# 	if anyone wants to continue using monocular data--please be aware of this
			sacbin <- binsacc(sacl$sac, sacr$sac)
			sacl <- sacbin$sac[,1:7]
			sacr <- sacbin$sac[,8:14]
			names(sacl) <- substr(names(sacl), 1, nchar(names(sacl))-2) # remove _L suffix
			names(sacr) <- substr(names(sacr), 1, nchar(names(sacr))-2) # remove _R suffix
			# what if we use averaged left and right eye instead?
			xyb <- DEG_PER_PIX * cbind(apply(data_trial[,c("xl","xr")], 1, mean), apply(data_trial[,c("yl","yr")], 1, mean) ) # mean position between eyes
			vb = vecvel(xyb, SAMPLING, VTYPE);       # computation of velocity components (x & y) for RIGHT eye data
			sacb = microsacc(xyb, vb, VTHR, MINDUR);  # averaged over eyes,output: '7 x number of saccades'-matrix
			sacb <- sacb$sac
			
			# (1) onset of saccade, (2) end of saccade, (3) peak velocity,
			# (4) horizontal component (dx = distance between saccade onset and offset) ,
			# (5) vertical component (dy)
			# (6) horizontal amplitude dX,
			# (7) vertical amplitude dY
			# amplitude is maximum expansion between saccade onset and offset
			
			for (sacdef in 1:length(sacdefLbls)) {#left, right, both
				whichsac <- get(whichsacs[sacdef])
				# whichsac = eval(whichsacs{sacdef});
				
				readsac <- whichsac
				if (nrow(readsac)==0) { # if no saccades detected
					#visualize_empty_sac(s,trial,xr,vr);   # visualize trial
					if (sacdef==2) { # lazy, do this only for RIGHT eye
						mis_collect <- rbind(mis_collect, c(s, trial, id))    # collect them
					}
				} else {
					
					# initialize cell arrays
					fixdur <- NULL          # for the current trial, initialize array for collection of fixation durations
					sacdur <- NULL          # for the current trial, initialize array for collection of fixation durations
					w  <- NULL              # ditto for word assignment
					l  <- NULL              # ditto for letter assignment
					lp  <- NULL             # ditto for character assignment (Chinese readers)
					chz  <- NULL            # ditto for character zone assignment (Chinese readers)
					fpos <- NULL            # ditto for MEAN fixation position
					fpos_a  <- NULL         # ditto for fixation position at the START of fixation
					fpos_b  <- NULL         # ditto for fixation position at the END of fixation
					a  <- NULL              # ditto for ONSET of fixation (relative to sentence onset)
					b  <- NULL              # ditto for OFFSET of fixation (relative to sentence onset)
					bad  <- NULL            # ditto for indexes of 'bad' fixation positions
					nsac_start <- rep(1, length(sacdefLbls))   # number of (valid) start sac./ fix. is set to 0
					nsac_end <- rep(1, length(sacdefLbls))   # 
					nsac <- nrow(readsac)   # number of saccades
					
					# get x-position of the eye [pixels], for the whole trial
					if(sacdefLbls[1]=="LEFT") {
						xpix <- data_trial$xl
					} else if(sacdefLbls[1]=="RIGHT") {
						xpix <- data_trial$xr
					} else {
						xpix <- apply(data_trial[,c("xl","xr")], 1, mean)
					}
					
					if (readsac[1,1] == 1) {
						# maybe I need to check sac(1,1) only --> start/stop of L and R appears to be always the same
						firstsac <- readsac[1,2]  # keep saccade end time
						readsac <- readsac[-1,]      # remove saccade ( Matlab: readsac(1,) <- [] )
						cat(sprintf('\tFirst saccade starts at time stamp 1 (trial: %i)!\n', trial))
					} else {
						firstsac <- 0
					}
					nsac_end[sacdef] <- nrow(readsac)   # nsac = number of (reading) saccades
					
					if (nrow(readsac) > 0) {
						for (k in 1:nsac_end[sacdef])  {          # use ALL saccades in order to determine fixations => for every saccade ...
							# ====================================================
							# 4)    Determine fixations
							# ====================================================
							if (k==1)  {
								# a(k) = 1;                 # (ALTE VERSION) special case of 1. fixation: there is no preceding saccade -> start = 1 
								a[k] <- firstsac + 1         # (Christiane) special case of 1. fixation: there is no preceding saccade -> start = 1
								b[k] <- readsac[k,1] - 1     # end of 1. fixation = start of 1. saccade
							} else {
								a[k] <- readsac[k-1,2] + 1   # a = start of fixation = end of preceding saccade (k-1)
								b[k] <- readsac[k,1] - 1     # b = end of k. fixation = start of k. saccade
							}
							fixdur[k] <- 1000/SAMPLING * (b[k] - a[k] + 1)    # computation of fixation duration
							sacdur[k] <- ( readsac[k,2] - readsac[k,1] + 1 ) * 1000/SAMPLING   # duration of following saccade
							# k now indexes FIXATIONS (instead of saccades)
						
							# ====================================================
							# 5)    Analysis of fixation location
							# ====================================================
						
							#TODO (jochen): here we sometimes have isempty(xpix(a(k):b(k))). Check why, and catch !! 
							# my guess: it's related to premature deletion of data in getMissingData
							thisFixX <- xpix[a[k]:b[k]]
						
							if ( length(thisFixX) == 0 ) {
								dummy <- NULL #dummy used for setting breakpoint
								print(paste('empty again for subject:', sub, 'trial:', trial))
							}
						
							fpos[k] <- mean(thisFixX); # MEAN fixation position
							# a = ONSET of fixation (time stamp relative to sentence onset)
							# b = OFFSET of fixation (time stamp relative to sentence onset)
						
							fpos_a[k] <- xpix[a[k]]  # position of the eyes at the START of fixation
							fpos_b[k] <- xpix[b[k]]  # position of the eyes at the END of fixation
						
							# word and letter assignment
							res <- wordLetterPos(fpos[k]-offset, sentenceWL, PIX_PER_CHAR)
							w[k] <- res$w
							l[k] <- res$l
						
						} # end of loop for k=1:nsac_end (fixation analysis)
					
						# ====================================================
						# 6)    Treatment of data left and/or right of sentence  
						# ====================================================
						# NB! the fixation location analysis always provides an assignment!
						# However, if the gaze was actually LEFT of the sentence, the fixation is assigned to the FIRST word (and first letter), and
						# if the gaze was actually RIGHT of the sentence, the fixation is assigned to the LAST word (last letter) of the sentence,
						# which apparently doesn't make sense. (we now treat this in wordLetterPos and assign NA (jl))
						# Therefore, it is implemented that 'fpos' must not be negative and must not be greater than the right sentence border
						# tolerance region: 1 letter
					
						excludeSent <- 0
						ixBad <- fpos < offset - PIX_PER_CHAR | fpos > rightSentenceBorder - offset + PIX_PER_CHAR 
						if (any(ixBad)) {
							bad <- which(ixBad)
							res <- treatBadData(bad, fpos, rightSentenceBorder - offset, PIX_PER_CHAR)
							nsac_start[sacdef] <- res$nsac_start
							nsac_end[sacdef] <- res$nsac_end
							excludeSent <- res$excludeSent
						}
					} else {excludeSent = 1}
					if (excludeSent == 1)  {  # if sentence is to be excluded
						if (sacdef == 2) {# lazy, do this only for RIGHT eye
							mis2_col_s_trial_id <- rbind(mis2_col_s_trial_id, c(s, trial, id))    # collect them
						}
						# subject id, trial nr., sentence nr.
					}
					
					if (excludeSent == 0) { # if sentence is not to be excluded
						# ====================================================
						# 7)    Write out fixation sequence
						# ====================================================
						
						for (k in nsac_start[sacdef]:nsac_end[sacdef]) { # for the first to last fixation ...
							if (k == 1) {
                                FirstLast <- 1      # marker for the FIRST fixation in a sentence
                            } else if (k == nsac_end[sacdef]) {
                                FirstLast <- 2  # marker for the LAST fixation in a sentence
                            } else {
                                FirstLast <- 0  # neither the first nor last fixation...
                            } 
							# resultStr <- sprintf('%i\t%i\t%i\t%i\t%f\t%i\t%i\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%i\n',  
							# 	s, id, nwords, w[k], l[k], fixdur[k], sacdur[k], 
							# 	readsac[k,"vpeak"], readsac[k,"dX"], readsac[k,"dY"], readsac[k,"dx"], readsac[k,"dy"], 
							# 	fpos_a[k], fpos_b[k], fpos[k], FirstLast)
							resultStr <- sprintf('%i\t%i\t%i\t%i\t%f\t%i\t%i\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%i\n',  
								subjID, id, nwords, w[k], l[k], fixdur[k], sacdur[k], 
								readsac[k,"vpeak"], readsac[k,"dX"], readsac[k,"dY"], readsac[k,"dx"], readsac[k,"dy"], 
								fpos_a[k], fpos_b[k], fpos[k], FirstLast)
								
							# cat(resultStr, file=outfilename[sacdefLbls[sacdef]], append = TRUE)
							cat(resultStr, file=outfiles[[sacdefLbls[sacdef]]], append = TRUE)
							flush(outfiles[[sacdefLbls[sacdef]]])
 						} # end of  k=nsac_start:nsac_end
					} # END OF: if res$excludeSent == 0
				} # END OF: if no saccades detected (nrow(readsac)!=0)
			} # for sacdef = 1:3 #left, right, both
		} # END OF: if no missing values
	} # END OF: for trial = 1:length(tab(:,1))
	
	# mis_col_s_trial_id has 3 columns: subject nr., trial nr., sentence id
	nrMissings <- 0
	if (length(mis_col_s_trial_id) > 0) {
		idx_mis <- (mis_col_s_trial_id[,1] == s)
		nrMissings <- sum(idx_mis);
	}
	mis_col_NrPerSubject = rbind(mis_col_NrPerSubject, c(s, nrMissings))
	# cat(sprintf('%i\t%i\n',s,nrMissings), file=fout_missings_b)
	
	# collect fixations that are left of the sentence
	# while they are not the first fixation in a sentence
	nrMissings2 <- 0
	if (length(mis2_col_s_trial_id) > 0) {
		idx_mis2 <- (mis2_col_s_trial_id[,1] == s)
		nrMissings2 <- sum(idx_mis2)
	}
	mis2_col_NrPerSubject <- rbind(mis2_col_NrPerSubject, c(s, nrMissings2))
	# cat(sprintf('%i\t%i\n',s,nrMissings2), file=fout_missings2_b)
} # for (s in 1:nos)


for (i in sacdefLbls) {
	close(outfiles[[i]])
}
#fclose(fout);  # close file for right-eye data

# fout_missings_b <- file(paste(missingpath, exp_str, '_missings_NrPerSubject.dat', sep=''), 'w')
   # write.table(mis2_col_NrPerSubject, file=fout_missings_b)
   # close(fout_missings_b);

missings_b <- paste(missingpath, '/', exp_str, '_missings_NrPerSubject.dat', sep='')
write.table(mis_col_NrPerSubject, file=)

missings2_b <- paste(missingpath, '/', exp_str, '_left_of_sent_and_not_first_NrPerSubject.dat', sep='')
write.table(mis2_col_NrPerSubject, file=missings2_b)

missings_a <- paste(missingpath, '/', exp_str, '_missings_s_trial_id.dat', sep='')
write.table(mis_col_s_trial_id, file=missings_a, row.names = FALSE, col.names = FALSE)

missings2_a <- paste(missingpath, '/', exp_str, '_left_of_sent_and_not_first_s_trial_id.dat', sep='')
write.table(mis2_col_s_trial_id, file=missings2_a, row.names = FALSE)

missings_all <- paste(missingpath, exp_str, '_all_excluded_sent_NrPerSubject.dat', sep='')
mis_all <- mis_col_NrPerSubject
mis_all[,2] = mis_all[,2] + mis2_col_NrPerSubject[,2]
write.table(mis_all, file=missings_all, row.names = FALSE)

cat('\n\tOk\n')

# diagnostics, for evaluating saccade detection parameters
# diagFixdur_l3
